# Copyright 2025 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


import pytest

from botocore.compat import HAS_CRT
from botocore.config import Config
from botocore.exceptions import FlexibleChecksumError
from tests import ClientHTTPStubber, patch_load_service_model
from tests.functional.test_useragent import (
    get_captured_ua_strings,
    parse_registered_feature_ids,
)

TEST_CHECKSUM_SERVICE_MODEL = {
    "version": "2.0",
    "documentation": "This is a test service.",
    "metadata": {
        "apiVersion": "2023-01-01",
        "endpointPrefix": "test",
        "protocol": "rest-json",
        "jsonVersion": "1.1",
        "serviceFullName": "Test Service",
        "serviceId": "Test Service",
        "signatureVersion": "v4",
        "signingName": "testservice",
        "uid": "testservice-2023-01-01",
    },
    "operations": {
        "HttpChecksumOperation": {
            "name": "HttpChecksumOperation",
            "http": {"method": "POST", "requestUri": "/HttpChecksumOperation"},
            "input": {"shape": "SomeInput"},
            "output": {"shape": "SomeOutput"},
            "httpChecksum": {
                "requestChecksumRequired": True,
                "requestAlgorithmMember": "checksumAlgorithm",
                "requestValidationModeMember": "validationMode",
                "responseAlgorithms": [
                    "CRC32",
                    "CRC32C",
                    "CRC64NVME",
                    "SHA1",
                    "SHA256",
                    "SHA512",
                    "MD5",
                    "XXHASH64",
                    "XXHASH3",
                    "XXHASH128",
                ],
            },
        },
        "HttpChecksumStreamingOperation": {
            "name": "HttpChecksumStreamingOperation",
            "http": {
                "method": "POST",
                "requestUri": "/HttpChecksumStreamingOperation",
            },
            "input": {"shape": "SomeStreamingInput"},
            "output": {"shape": "SomeStreamingOutput"},
            "httpChecksum": {
                "requestChecksumRequired": True,
                "requestAlgorithmMember": "checksumAlgorithm",
                "requestValidationModeMember": "validationMode",
                "responseAlgorithms": [
                    "CRC32",
                    "CRC32C",
                    "CRC64NVME",
                    "SHA1",
                    "SHA256",
                    "SHA512",
                    "MD5",
                    "XXHASH64",
                    "XXHASH3",
                    "XXHASH128",
                ],
            },
        },
    },
    "shapes": {
        "ChecksumAlgorithm": {
            "type": "string",
            "enum": {
                "CRC32",
                "CRC32C",
                "CRC64NVME",
                "SHA1",
                "SHA256",
                "SHA512",
                "MD5",
                "XXHASH64",
                "XXHASH3",
                "XXHASH128",
            },
            "member": {"shape": "MockOpParam"},
        },
        "ValidationMode": {"type": "string", "enum": {"ENABLE"}},
        "String": {"type": "string"},
        "Blob": {"type": "blob"},
        "SomeStreamingOutput": {
            "type": "structure",
            "members": {"body": {"shape": "Blob", "streaming": True}},
            "payload": "body",
        },
        "SomeStreamingInput": {
            "type": "structure",
            "required": ["body"],
            "members": {
                "body": {
                    "shape": "Blob",
                    "streaming": True,
                },
                "validationMode": {
                    "shape": "ValidationMode",
                    "location": "header",
                    "locationName": "x-amz-response-validation-mode",
                },
                "checksumAlgorithm": {
                    "shape": "ChecksumAlgorithm",
                    "location": "header",
                    "locationName": "x-amz-request-algorithm",
                },
            },
            "payload": "body",
        },
        "SomeInput": {
            "type": "structure",
            "required": ["body"],
            "members": {
                "body": {"shape": "String"},
                "validationMode": {
                    "shape": "ValidationMode",
                    "location": "header",
                    "locationName": "x-amz-response-validation-mode",
                },
                "checksumAlgorithm": {
                    "shape": "ChecksumAlgorithm",
                    "location": "header",
                    "locationName": "x-amz-request-algorithm",
                },
                "ChecksumMD5": {
                    "shape": "ChecksumMD5",
                    "location": "header",
                    "locationName": "x-amz-checksum-md5",
                },
            },
            "payload": "body",
        },
        "SomeOutput": {
            "type": "structure",
        },
        "ChecksumMD5": {
            "type": "string",
        },
    },
}

TEST_CHECKSUM_RULESET = {
    "version": "1.0",
    "parameters": {},
    "rules": [
        {
            "conditions": [],
            "type": "endpoint",
            "endpoint": {
                "url": "https://foo.bar",
                "properties": {},
                "headers": {},
            },
        }
    ],
}


def setup_test_client(patched_session, monkeypatch, config=None):
    patch_load_service_model(
        patched_session,
        monkeypatch,
        TEST_CHECKSUM_SERVICE_MODEL,
        TEST_CHECKSUM_RULESET,
    )
    return patched_session.create_client(
        "testservice", region_name="us-west-2", config=config
    )


def _request_checksum_calculation_cases():
    request_payload = "Hello world"
    cases = [
        (
            None,
            request_payload,
            {
                "x-amz-request-algorithm": "CRC32",
                "x-amz-checksum-crc32": "i9aeUg==",
            },
        ),
        (
            "CRC32",
            request_payload,
            {
                "x-amz-request-algorithm": "CRC32",
                "x-amz-checksum-crc32": "i9aeUg==",
            },
        ),
        (
            "SHA1",
            request_payload,
            {
                "x-amz-request-algorithm": "SHA1",
                "x-amz-checksum-sha1": "e1AsOh9IyGCa4hLN+2Od7jlnP14=",
            },
        ),
        (
            "SHA256",
            request_payload,
            {
                "x-amz-request-algorithm": "SHA256",
                "x-amz-checksum-sha256": "ZOyIygCyaOW6GjVnihtTFtIS9PNmskdyMlNKiuyjfzw=",
            },
        ),
        (
            "SHA512",
            request_payload,
            {
                "x-amz-request-algorithm": "SHA512",
                "x-amz-checksum-sha512": "t/eDuu2Cl/DbkXRiGE/08I5pwtXl95qUJgD5cl9Yzh8pwYE5v4CwbA//K900c4RS7PQMSIwip+PYDN9vnBwNRw==",
            },
        ),
    ]
    if HAS_CRT:
        cases.extend(
            [
                (
                    "CRC32C",
                    request_payload,
                    {
                        "x-amz-request-algorithm": "CRC32C",
                        "x-amz-checksum-crc32c": "crUfeA==",
                    },
                ),
                (
                    "CRC64NVME",
                    request_payload,
                    {
                        "x-amz-request-algorithm": "CRC64NVME",
                        "x-amz-checksum-crc64nvme": "OOJZ0D8xKts=",
                    },
                ),
                (
                    "XXHASH64",
                    request_payload,
                    {
                        "x-amz-request-algorithm": "XXHASH64",
                        "x-amz-checksum-xxhash64": "xQCwyRKzdtg=",
                    },
                ),
                (
                    "XXHASH3",
                    request_payload,
                    {
                        "x-amz-request-algorithm": "XXHASH3",
                        "x-amz-checksum-xxhash3": "tqy52Eo4/3Q=",
                    },
                ),
                (
                    "XXHASH128",
                    request_payload,
                    {
                        "x-amz-request-algorithm": "XXHASH128",
                        "x-amz-checksum-xxhash128": "c1H4mBL5c4K5HQWzHgTdfw==",
                    },
                ),
            ]
        )
    return cases


@pytest.mark.parametrize(
    "checksum_algorithm, request_payload, expected_headers",
    _request_checksum_calculation_cases(),
)
def test_request_checksum_calculation(
    patched_session,
    monkeypatch,
    checksum_algorithm,
    request_payload,
    expected_headers,
):
    client = setup_test_client(patched_session, monkeypatch)

    with ClientHTTPStubber(client, strict=True) as http_stubber:
        http_stubber.add_response(status=200, body=b"<response/>")
        operation_kwargs = {"body": request_payload}
        if checksum_algorithm:
            operation_kwargs["checksumAlgorithm"] = checksum_algorithm
        client.http_checksum_operation(**operation_kwargs)
        actual_headers = http_stubber.requests[0].headers
        for key, val in expected_headers.items():
            assert key in actual_headers
            assert actual_headers[key].decode() == val


def _streaming_request_checksum_calculation_cases():
    request_payload = "Hello world"
    cases = [
        (
            None,
            request_payload,
            {
                "x-amz-request-algorithm": "CRC32",
                "content-encoding": "aws-chunked",
                "x-amz-trailer": "x-amz-checksum-crc32",
            },
            {"x-amz-checksum-crc32": "i9aeUg=="},
        ),
        (
            "CRC32",
            request_payload,
            {
                "x-amz-request-algorithm": "CRC32",
                "content-encoding": "aws-chunked",
                "x-amz-trailer": "x-amz-checksum-crc32",
            },
            {"x-amz-checksum-crc32": "i9aeUg=="},
        ),
        (
            "SHA1",
            request_payload,
            {
                "x-amz-request-algorithm": "SHA1",
                "content-encoding": "aws-chunked",
                "x-amz-trailer": "x-amz-checksum-sha1",
            },
            {"x-amz-checksum-sha1": "e1AsOh9IyGCa4hLN+2Od7jlnP14="},
        ),
        (
            "SHA256",
            request_payload,
            {
                "x-amz-request-algorithm": "SHA256",
                "content-encoding": "aws-chunked",
                "x-amz-trailer": "x-amz-checksum-sha256",
            },
            {
                "x-amz-checksum-sha256": "ZOyIygCyaOW6GjVnihtTFtIS9PNmskdyMlNKiuyjfzw="
            },
        ),
        (
            "SHA512",
            request_payload,
            {
                "x-amz-request-algorithm": "SHA512",
                "content-encoding": "aws-chunked",
                "x-amz-trailer": "x-amz-checksum-sha512",
            },
            {
                "x-amz-checksum-sha512": "t/eDuu2Cl/DbkXRiGE/08I5pwtXl95qUJgD5cl9Yzh8pwYE5v4CwbA//K900c4RS7PQMSIwip+PYDN9vnBwNRw=="
            },
        ),
    ]
    if HAS_CRT:
        cases.extend(
            [
                (
                    "CRC32C",
                    request_payload,
                    {
                        "x-amz-request-algorithm": "CRC32C",
                        "content-encoding": "aws-chunked",
                        "x-amz-trailer": "x-amz-checksum-crc32c",
                    },
                    {"x-amz-checksum-crc32c": "crUfeA=="},
                ),
                (
                    "CRC64NVME",
                    request_payload,
                    {
                        "x-amz-request-algorithm": "CRC64NVME",
                        "content-encoding": "aws-chunked",
                        "x-amz-trailer": "x-amz-checksum-crc64nvme",
                    },
                    {"x-amz-checksum-crc64nvme": "OOJZ0D8xKts="},
                ),
                (
                    "XXHASH64",
                    request_payload,
                    {
                        "x-amz-request-algorithm": "XXHASH64",
                        "content-encoding": "aws-chunked",
                        "x-amz-trailer": "x-amz-checksum-xxhash64",
                    },
                    {"x-amz-checksum-xxhash64": "xQCwyRKzdtg="},
                ),
                (
                    "XXHASH3",
                    request_payload,
                    {
                        "x-amz-request-algorithm": "XXHASH3",
                        "content-encoding": "aws-chunked",
                        "x-amz-trailer": "x-amz-checksum-xxhash3",
                    },
                    {"x-amz-checksum-xxhash3": "tqy52Eo4/3Q="},
                ),
                (
                    "XXHASH128",
                    request_payload,
                    {
                        "x-amz-request-algorithm": "XXHASH128",
                        "content-encoding": "aws-chunked",
                        "x-amz-trailer": "x-amz-checksum-xxhash128",
                    },
                    {"x-amz-checksum-xxhash128": "c1H4mBL5c4K5HQWzHgTdfw=="},
                ),
            ]
        )
    return cases


@pytest.mark.parametrize(
    "checksum_algorithm, request_payload, expected_headers, expected_trailers",
    _streaming_request_checksum_calculation_cases(),
)
def test_streaming_request_checksum_calculation(
    patched_session,
    monkeypatch,
    checksum_algorithm,
    request_payload,
    expected_headers,
    expected_trailers,
):
    client = setup_test_client(patched_session, monkeypatch)

    with ClientHTTPStubber(client, strict=True) as http_stubber:
        http_stubber.add_response(status=200, body=b"<response/>")
        operation_kwargs = {"body": request_payload}
        if checksum_algorithm:
            operation_kwargs["checksumAlgorithm"] = checksum_algorithm
        client.http_checksum_streaming_operation(**operation_kwargs)
        request = http_stubber.requests[0]
        actual_headers = request.headers
        for key, val in expected_headers.items():
            assert key in actual_headers
            assert actual_headers[key].decode() == val
        read_body = request.body.read()
        for key, val in expected_trailers.items():
            assert f"{key}:{val}".encode() in read_body


def _successful_response_checksum_validation_cases():
    response_payload = "Hello world"
    cases = [
        (
            "CRC32",
            response_payload,
            {"x-amz-checksum-crc32": "i9aeUg=="},
        ),
        (
            "SHA1",
            response_payload,
            {"x-amz-checksum-sha1": "e1AsOh9IyGCa4hLN+2Od7jlnP14="},
        ),
        (
            "SHA256",
            response_payload,
            {
                "x-amz-checksum-sha256": "ZOyIygCyaOW6GjVnihtTFtIS9PNmskdyMlNKiuyjfzw="
            },
        ),
        (
            "SHA512",
            response_payload,
            {
                "x-amz-checksum-sha512": "t/eDuu2Cl/DbkXRiGE/08I5pwtXl95qUJgD5cl9Yzh8pwYE5v4CwbA//K900c4RS7PQMSIwip+PYDN9vnBwNRw=="
            },
        ),
    ]
    if HAS_CRT:
        cases.extend(
            [
                (
                    "CRC32C",
                    response_payload,
                    {"x-amz-checksum-crc32c": "crUfeA=="},
                ),
                (
                    "CRC64NVME",
                    response_payload,
                    {"x-amz-checksum-crc64nvme": "OOJZ0D8xKts="},
                ),
                (
                    "XXHASH64",
                    response_payload,
                    {"x-amz-checksum-xxhash64": "xQCwyRKzdtg="},
                ),
                (
                    "XXHASH3",
                    response_payload,
                    {"x-amz-checksum-xxhash3": "tqy52Eo4/3Q="},
                ),
                (
                    "XXHASH128",
                    response_payload,
                    {"x-amz-checksum-xxhash128": "c1H4mBL5c4K5HQWzHgTdfw=="},
                ),
            ]
        )
    return cases


@pytest.mark.parametrize(
    "checksum_algorithm, response_payload, response_headers",
    _successful_response_checksum_validation_cases(),
)
def test_successful_response_checksum_validation(
    patched_session,
    monkeypatch,
    checksum_algorithm,
    response_payload,
    response_headers,
):
    client = setup_test_client(patched_session, monkeypatch)

    with ClientHTTPStubber(client, strict=True) as http_stubber:
        http_stubber.add_response(
            status=200,
            body=response_payload.encode(),
            headers=response_headers,
        )
        operation_kwargs = {
            "body": response_payload,
            "checksumAlgorithm": checksum_algorithm,
        }

        client.http_checksum_operation(**operation_kwargs)


@pytest.mark.parametrize(
    "checksum_algorithm, response_payload, response_headers",
    _successful_response_checksum_validation_cases(),
)
def test_successful_streaming_response_checksum_validation(
    patched_session,
    monkeypatch,
    checksum_algorithm,
    response_payload,
    response_headers,
):
    client = setup_test_client(patched_session, monkeypatch)

    with ClientHTTPStubber(client, strict=True) as http_stubber:
        http_stubber.add_response(
            status=200,
            body=response_payload.encode(),
            headers=response_headers,
        )
        response = client.http_checksum_streaming_operation(
            body=response_payload,
            checksumAlgorithm=checksum_algorithm,
        )
        response["body"].read()


def _unsuccessful_response_checksum_validation_cases():
    response_payload = "Hello world"
    cases = [
        (
            "CRC32",
            response_payload,
            {"x-amz-checksum-crc32": "bm90LWEtY2hlY2tzdW0="},
            "i9aeUg==",
        ),
        (
            "SHA1",
            response_payload,
            {"x-amz-checksum-sha1": "bm90LWEtY2hlY2tzdW0="},
            "e1AsOh9IyGCa4hLN+2Od7jlnP14=",
        ),
        (
            "SHA256",
            response_payload,
            {"x-amz-checksum-sha256": "bm90LWEtY2hlY2tzdW0="},
            "ZOyIygCyaOW6GjVnihtTFtIS9PNmskdyMlNKiuyjfzw=",
        ),
        (
            "SHA512",
            response_payload,
            {"x-amz-checksum-sha512": "bm90LWEtY2hlY2tzdW0="},
            "t/eDuu2Cl/DbkXRiGE/08I5pwtXl95qUJgD5cl9Yzh8pwYE5v4CwbA//K900c4RS7PQMSIwip+PYDN9vnBwNRw==",
        ),
    ]
    if HAS_CRT:
        cases.extend(
            [
                (
                    "CRC32C",
                    response_payload,
                    {"x-amz-checksum-crc32c": "bm90LWEtY2hlY2tzdW0="},
                    "crUfeA==",
                ),
                (
                    "CRC64NVME",
                    response_payload,
                    {"x-amz-checksum-crc64nvme": "bm90LWEtY2hlY2tzdW0="},
                    "OOJZ0D8xKts=",
                ),
                (
                    "XXHASH64",
                    response_payload,
                    {"x-amz-checksum-xxhash64": "bm90LWEtY2hlY2tzdW0="},
                    "xQCwyRKzdtg=",
                ),
                (
                    "XXHASH3",
                    response_payload,
                    {"x-amz-checksum-xxhash3": "bm90LWEtY2hlY2tzdW0="},
                    "tqy52Eo4/3Q=",
                ),
                (
                    "XXHASH128",
                    response_payload,
                    {"x-amz-checksum-xxhash128": "bm90LWEtY2hlY2tzdW0="},
                    "c1H4mBL5c4K5HQWzHgTdfw==",
                ),
            ]
        )
    return cases


@pytest.mark.parametrize(
    "checksum_algorithm, response_payload, response_headers, expected_checksum",
    _unsuccessful_response_checksum_validation_cases(),
)
def test_unsuccessful_response_checksum_validation(
    patched_session,
    monkeypatch,
    checksum_algorithm,
    response_payload,
    response_headers,
    expected_checksum,
):
    client = setup_test_client(patched_session, monkeypatch)

    with ClientHTTPStubber(client, strict=True) as http_stubber:
        http_stubber.add_response(
            status=200,
            body=response_payload.encode(),
            headers=response_headers,
        )
        operation_kwargs = {
            "body": response_payload,
            "checksumAlgorithm": checksum_algorithm,
        }
        with pytest.raises(FlexibleChecksumError) as expected_error:
            client.http_checksum_operation(**operation_kwargs)
        error_msg = "Expected checksum {} did not match calculated checksum: {}".format(
            response_headers[f'x-amz-checksum-{checksum_algorithm.lower()}'],
            expected_checksum,
        )
        assert str(expected_error.value) == error_msg


@pytest.mark.parametrize(
    "checksum_algorithm, response_payload, response_headers, expected_checksum",
    _unsuccessful_response_checksum_validation_cases(),
)
def test_unsuccessful_streaming_response_checksum_validation(
    patched_session,
    monkeypatch,
    checksum_algorithm,
    response_payload,
    response_headers,
    expected_checksum,
):
    client = setup_test_client(patched_session, monkeypatch)

    with ClientHTTPStubber(client, strict=True) as http_stubber:
        http_stubber.add_response(
            status=200,
            body=response_payload.encode(),
            headers=response_headers,
        )
        response = client.http_checksum_streaming_operation(
            body=response_payload,
            checksumAlgorithm=checksum_algorithm,
        )
        with pytest.raises(FlexibleChecksumError) as expected_error:
            response["body"].read()
        error_msg = "Expected checksum {} did not match calculated checksum: {}".format(
            response_headers[f'x-amz-checksum-{checksum_algorithm.lower()}'],
            expected_checksum,
        )
        assert str(expected_error.value) == error_msg


def _checksum_user_agent_feature_id_cases():
    request_payload = "Hello world"
    cases = [
        (None, None, None, request_payload, ["U", "b", "Z"]),
        (
            "CRC32",
            "when_required",
            "when_required",
            request_payload,
            ["U", "a", "c"],
        ),
        (
            "SHA1",
            "when_supported",
            "when_supported",
            request_payload,
            ["X", "Z", "b"],
        ),
        (
            "SHA256",
            "when_supported",
            "when_required",
            request_payload,
            ["Y", "Z", "c"],
        ),
        (
            "SHA1",
            "when_required",
            "when_supported",
            request_payload,
            ["X", "a", "b"],
        ),
        (
            "SHA512",
            None,
            None,
            request_payload,
            ["AF", "Z", "b"],
        ),
    ]
    if HAS_CRT:
        cases.extend(
            [("CRC32C", None, None, request_payload, ["V", "Z", 'b'])]
        )
        cases.extend(
            [
                (
                    "CRC64NVME",
                    None,
                    "when_required",
                    request_payload,
                    ["W", "Z", "c"],
                )
            ]
        )
        cases.extend(
            [
                (
                    "XXHASH3",
                    None,
                    None,
                    request_payload,
                    ["AG", "Z", "b"],
                )
            ]
        )
        cases.extend(
            [
                (
                    "XXHASH64",
                    None,
                    None,
                    request_payload,
                    ["AH", "Z", "b"],
                )
            ]
        )
        cases.extend(
            [
                (
                    "XXHASH128",
                    None,
                    None,
                    request_payload,
                    ["AI", "Z", "b"],
                )
            ]
        )
    return cases


@pytest.mark.parametrize(
    "checksum_algorithm, checksum_calculation, checksum_validation, request_payload, expected_feature_ids",
    _checksum_user_agent_feature_id_cases(),
)
def test_user_agent_has_checksum_request_feature_id(
    patched_session,
    monkeypatch,
    checksum_algorithm,
    checksum_calculation,
    checksum_validation,
    request_payload,
    expected_feature_ids,
):
    client = setup_test_client(
        patched_session,
        monkeypatch,
        config=Config(
            request_checksum_calculation=checksum_calculation,
            response_checksum_validation=checksum_validation,
        ),
    )

    with ClientHTTPStubber(client, strict=True) as http_stubber:
        http_stubber.add_response(
            status=200,
            body=b"<response/>",
        )
        operation_kwargs = {"body": request_payload}
        if checksum_algorithm:
            operation_kwargs["checksumAlgorithm"] = checksum_algorithm
        client.http_checksum_streaming_operation(**operation_kwargs)
    ua_string = get_captured_ua_strings(http_stubber)[0]
    feature_list = parse_registered_feature_ids(ua_string)
    for feature_id in expected_feature_ids:
        assert feature_id in feature_list


def test_user_agent_has_md5_feature_id(
    patched_session,
    monkeypatch,
):
    """Test that ChecksumMD5 parameter usage registers the AE feature ID in user agent.
    This test is not part of test_user_agent_has_checksum_request_feature_id since automatic
    MD5 calculation is not supported"""

    client = setup_test_client(
        patched_session,
        monkeypatch,
    )

    with ClientHTTPStubber(client, strict=True) as http_stubber:
        http_stubber.add_response(
            status=200,
            body=b"<response/>",
        )
        client.http_checksum_operation(
            body="test", ChecksumMD5="rL0Y20zC+Fzt72VPzMSk2A=="
        )

    ua_string = get_captured_ua_strings(http_stubber)[0]
    feature_list = parse_registered_feature_ids(ua_string)

    assert "AE" in feature_list



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from botocore import xform_name
from botocore.docs.client import ClientContextParamsDocumenter
from botocore.docs.service import ServiceDocumenter
from tests.functional.docs import BaseDocsFunctionalTest


class TestS3Docs(BaseDocsFunctionalTest):
    def test_auto_populates_sse_customer_key_md5(self):
        self.assert_is_documented_as_autopopulated_param(
            service_name='s3',
            method_name='put_object',
            param_name='SSECustomerKeyMD5',
        )

    def test_auto_populates_copy_source_sse_customer_key_md5(self):
        self.assert_is_documented_as_autopopulated_param(
            service_name='s3',
            method_name='copy_object',
            param_name='CopySourceSSECustomerKeyMD5',
        )

    def test_hides_content_md5_when_impossible_to_provide(self):
        modified_methods = [
            'delete_objects',
            'put_bucket_acl',
            'put_bucket_cors',
            'put_bucket_lifecycle',
            'put_bucket_logging',
            'put_bucket_policy',
            'put_bucket_notification',
            'put_bucket_tagging',
            'put_bucket_replication',
            'put_bucket_website',
            'put_bucket_request_payment',
            'put_object_acl',
            'put_bucket_versioning',
        ]
        ServiceDocumenter(
            's3', self._session, self.root_services_path
        ).document_service()
        for method_name in modified_methods:
            contents = self.get_client_method_contents('s3', method_name)
            method_contents = self.get_method_document_block(
                method_name, contents
            )
            self.assertNotIn(
                'ContentMD5=\'string\'', method_contents.decode('utf-8')
            )

    def test_generate_presigned_url_documented(self):
        content = self.get_docstring_for_method('s3', 'generate_presigned_url')
        self.assert_contains_line('generate_presigned_url', content)

    def test_copy_source_documented_as_union_type(self):
        content = self.get_docstring_for_method('s3', 'copy_object')
        dict_form = (
            "{'Bucket': 'string', 'Key': 'string', 'VersionId': 'string'}"
        )
        self.assert_contains_line(
            f"CopySource='string' or {dict_form}", content
        )

    def test_copy_source_param_docs_also_modified(self):
        content = self.get_docstring_for_method('s3', 'copy_object')
        param_docs = self.get_parameter_document_block('CopySource', content)
        # We don't want to overspecify the test, so I've picked
        # an arbitrary line from the customized docs.
        self.assert_contains_line(
            "You can also provide this value as a dictionary", param_docs
        )

    def test_s3_context_params_omitted(self):
        omitted_params = ClientContextParamsDocumenter.OMITTED_CONTEXT_PARAMS
        s3_omitted_params = omitted_params['s3']
        content = ServiceDocumenter(
            's3', self._session, self.root_services_path
        ).document_service()
        for param in s3_omitted_params:
            param_name = f'``{xform_name(param)}``'
            self.assert_not_contains_line(param_name, content)

    def test_s3control_context_params_omitted(self):
        omitted_params = ClientContextParamsDocumenter.OMITTED_CONTEXT_PARAMS
        s3control_omitted_params = omitted_params['s3control']
        content = ServiceDocumenter(
            's3control', self._session, self.root_services_path
        ).document_service()
        for param in s3control_omitted_params:
            param_name = f'``{xform_name(param)}``'
            self.assert_not_contains_line(param_name, content)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import botocore.session
from tests import unittest


class TestRoute53Pagination(unittest.TestCase):
    def setUp(self):
        self.session = botocore.session.get_session()
        self.client = self.session.create_client('route53', 'us-west-2')

    # def test_paginate_with_max_items(self):
    #     # Route53 has a string type for MaxItems.  We need to ensure that this
    #     # still works without any issues.
    #     paginator = self.client.get_paginator('list_hosted_zones')
    #     results = list(paginator.paginate(PaginationConfig={'MaxItems': '1'}))
    #     self.assertTrue(len(results) >= 0)

    # def test_paginate_with_deprecated_paginator_and_limited_input_tokens(self):
    #     paginator = self.client.get_paginator('list_resource_record_sets')

    #     # We're making sure the paginator gets set without failing locally, so
    #     # a ClientError is acceptable. In this case, the Hosted Zone specified
    #     # does not exist.
    #     with self.assertRaises(ClientError):
    #         results = list(
    #             paginator.paginate(
    #                 PaginationConfig={
    #                     'MaxItems': '1',
    #                     'StartingToken': 'my.domain.name.',
    #                 },
    #                 HostedZoneId="foo",
    #             )
    #         )
    #         self.assertTrue(len(results) >= 0)


if __name__ == '__main__':
    unittest.main()



# Copyright 2026 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest

from botocore import xform_name
from botocore.session import get_session
from botocore.stub import Stubber

_OTEL_OPERATIONS = [
    'GetOTelEnrichment',
    'StartOTelEnrichment',
    'StopOTelEnrichment',
]


@pytest.mark.validates_models
@pytest.mark.parametrize("operation", _OTEL_OPERATIONS)
def test_otel_xform_name(operation):
    assert 'otel' in xform_name(operation, '_')
    assert 'otel' in xform_name(operation, '-')


class TestCloudWatchOTelEnrichment:
    def setup_method(self):
        session = get_session()
        self.client = session.create_client('cloudwatch', 'us-east-1')
        self.stubber = Stubber(self.client)
        self.stubber.activate()

    @pytest.mark.parametrize(
        "old_name, new_name, response",
        [
            (
                'get_o_tel_enrichment',
                'get_otel_enrichment',
                {'Status': 'Running'},
            ),
            ('start_o_tel_enrichment', 'start_otel_enrichment', {}),
            ('stop_o_tel_enrichment', 'stop_otel_enrichment', {}),
        ],
    )
    def test_otel_enrichment_aliased(self, old_name, new_name, response):
        self.stubber.add_response(new_name, response)
        self.stubber.add_response(new_name, response)

        getattr(self.client, old_name)()
        getattr(self.client, new_name)()

        self.stubber.assert_no_pending_responses()



# Copyright 2016 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
import pytest

import botocore.docs.bcdoc.docstringparser as parser
from botocore.docs.bcdoc.restdoc import ReSTDocument
from tests import mock, unittest


class TestDocStringParser(unittest.TestCase):
    def parse(self, html):
        docstring_parser = parser.DocStringParser(ReSTDocument())
        docstring_parser.feed(html)
        docstring_parser.close()
        return docstring_parser.doc.getvalue()

    def assert_contains_exact_lines_in_order(self, actual, expected):
        # Get each line and filter out empty lines
        contents = actual.split(b'\n')
        contents = [line for line in contents if line and not line.isspace()]

        for line in expected:
            self.assertIn(line, contents)
            beginning = contents.index(line)
            contents = contents[beginning:]

    def test_tag_with_collapsible_spaces(self):
        html = "<p>  a       bcd efg </p>"
        result = self.parse(html)
        self.assert_contains_exact_lines_in_order(result, [b'a bcd efg'])

    def test_nested_lists(self):
        html = "<ul><li>Wello</li><ul><li>Horld</li></ul></ul>"
        result = self.parse(html)
        self.assert_contains_exact_lines_in_order(
            result, [b'* Wello', b'  * Horld']
        )

    def test_nested_lists_with_extra_white_space(self):
        html = "<ul> <li> Wello</li><ul> <li> Horld</li></ul></ul>"
        result = self.parse(html)
        self.assert_contains_exact_lines_in_order(
            result, [b'* Wello', b'  * Horld']
        )

    def test_link_with_no_period(self):
        html = "<p>This is a test <a href='https://testing.com'>Link</a></p>"
        result = self.parse(html)
        self.assert_contains_exact_lines_in_order(
            result, [b'This is a test `Link <https://testing.com>`__']
        )

    def test_link_with_period(self):
        html = "<p>This is a test <a href='https://testing.com'>Link</a>.</p>"
        result = self.parse(html)
        self.assert_contains_exact_lines_in_order(
            result, [b'This is a test `Link <https://testing.com>`__.']
        )

    def test_code_with_empty_link(self):
        html = "<p>Foo <code> <a>Link</a> </code></p>"
        result = self.parse(html)
        self.assert_contains_exact_lines_in_order(result, [b'Foo ``Link``'])

    def test_code_with_link_spaces(self):
        html = "<p>Foo <code> <a href=\"https://aws.dev\">Link</a> </code></p>"
        result = self.parse(html)
        self.assert_contains_exact_lines_in_order(result, [b'Foo ``Link``'])

    def test_code_with_link_no_spaces(self):
        html = "<p>Foo <code><a href=\"https://aws.dev\">Link</a></code></p>"
        result = self.parse(html)
        self.assert_contains_exact_lines_in_order(result, [b'Foo ``Link``'])

    def test_href_with_spaces(self):
        html = "<p><a href=\" https://testing.com\">Link</a></p>"
        result = self.parse(html)
        self.assert_contains_exact_lines_in_order(
            result, [b' `Link <https://testing.com>`__']
        )

    def test_bold_with_nested_formatting(self):
        html = "<b><code>Test</code>test<a href=\" https://testing.com\">Link</a></b>"
        result = self.parse(html)
        self.assert_contains_exact_lines_in_order(
            result, [b'``Test``test `Link <https://testing.com>`__']
        )

    def test_link_with_nested_formatting(self):
        html = "<a href=\"https://testing.com\"><code>Test</code></a>"
        result = self.parse(html)
        self.assert_contains_exact_lines_in_order(
            result, [b'`Test <https://testing.com>`__']
        )

    def test_indentation_with_spaces_between_tags(self):
        # Leading spaces inserted as padding between HTML tags can lead to
        # unexpected indentation in RST. For example, consider the space between
        # ``<p>`` and ``<b>`` in the third line of this example:
        html = (
            "<p>First paragraph </p> "
            "<note> <p> Second paragraph in note </p> </note> "
            "<p> <b>Bold statement:</b> Third paragraph</p>"
            "<p>Last paragraph</p> "
        )
        # If kept, it will appear as indentation in RST and have the effect of
        # pulling the third paragraph into the note:
        #
        # ```
        # First paragraph
        #
        # ..note::
        #   Second paragraph in note <-- intentionally indented
        #
        #  **Bold statement:** Third paragraph <-- unintentionally indented
        #
        # Last paragraph <-- first non-indented paragraph after note
        # ```
        result = self.parse(html)
        self.assert_contains_exact_lines_in_order(
            result,
            #  ↓ no whitespace here
            [b'**Bold statement:** Third paragraph'],
        )


class TestHTMLTree(unittest.TestCase):
    def setUp(self):
        self.style = mock.Mock()
        self.doc = mock.Mock()
        self.doc.style = self.style
        self.tree = parser.HTMLTree(self.doc)

    def test_add_tag(self):
        self.tree.add_tag('foo')
        self.assertIsInstance(self.tree.current_node, parser.TagNode)
        self.assertEqual(self.tree.current_node.tag, 'foo')

    def test_add_unsupported_tag(self):
        del self.style.start_foo
        del self.style.end_foo
        self.tree.add_tag('foo')
        self.assertIn('foo', self.tree.unhandled_tags)

    def test_add_data(self):
        self.tree.add_data('foo')
        self.assertNotIsInstance(self.tree.current_node, parser.DataNode)
        node = self.tree.head.children[0]
        self.assertIsInstance(node, parser.DataNode)
        self.assertEqual(node.data, 'foo')


class TestStemNode(unittest.TestCase):
    def setUp(self):
        self.style = mock.Mock()
        self.doc = mock.Mock()
        self.doc.style = self.style
        self.node = parser.StemNode()

    def test_add_child(self):
        child = parser.StemNode()
        self.node.add_child(child)
        self.assertIn(child, self.node.children)
        self.assertEqual(child.parent, self.node)

    def test_write(self):
        self.node.add_child(mock.Mock())
        self.node.add_child(mock.Mock())

        self.node.write(mock.Mock())
        for child in self.node.children:
            self.assertTrue(child.write.called)


class TestTagNode(unittest.TestCase):
    def setUp(self):
        self.style = mock.Mock()
        self.doc = mock.Mock()
        self.doc.style = self.style
        self.tag = 'foo'
        self.node = parser.TagNode(self.tag)

    def test_write_calls_style(self):
        self.node.write(self.doc)
        self.assertTrue(self.style.start_foo.called)
        self.assertTrue(self.style.end_foo.called)

    def test_write_unsupported_tag(self):
        del self.style.start_foo
        del self.style.end_foo

        try:
            self.node.write(self.doc)
        except AttributeError as e:
            self.fail(str(e))


class TestDataNode(unittest.TestCase):
    def setUp(self):
        self.style = mock.Mock()
        self.doc = mock.Mock()
        self.doc.style = self.style
        self.doc.translate_words.return_value = []

    def test_string_data(self):
        node = parser.DataNode('foo')
        self.assertEqual(node.data, 'foo')

    def test_non_string_data_raises_error(self):
        with self.assertRaises(ValueError):
            parser.DataNode(5)

    def test_lstrip(self):
        node = parser.DataNode(' foo')
        node.lstrip()
        self.assertEqual(node.data, 'foo')

    def test_write(self):
        node = parser.DataNode('foo   bar baz')
        self.doc.translate_words.return_value = ['foo', 'bar', 'baz']
        node.write(self.doc)
        self.doc.handle_data.assert_called_once_with('foo bar baz')

    def test_write_empty_string(self):
        node = parser.DataNode('')
        node.write(self.doc)
        self.assertFalse(self.doc.handle_data.called)


@pytest.mark.parametrize(
    'data, lstrip, rstrip, both',
    [
        ('foo', 'foo', 'foo', 'foo'),
        (' foo', 'foo', ' foo', 'foo'),
        ('   foo', 'foo', '   foo', 'foo'),
        ('\tfoo', 'foo', '\tfoo', 'foo'),
        ('\t \t foo', 'foo', '\t \t foo', 'foo'),
        ('foo ', 'foo ', 'foo', 'foo'),
        ('foo  ', 'foo  ', 'foo', 'foo'),
        ('foo\t\t', 'foo\t\t', 'foo', 'foo'),
    ],
)
def test_datanode_stripping(data, lstrip, rstrip, both):
    doc = mock.Mock()
    doc.style = mock.Mock()
    doc.translate_words.side_effect = lambda words: words

    node = parser.DataNode(data)
    node.lstrip()
    node.write(doc)
    doc.handle_data.assert_called_once_with(lstrip)
    doc.handle_data.reset_mock()

    node = parser.DataNode(data)
    node.rstrip()
    node.write(doc)
    doc.handle_data.assert_called_once_with(rstrip)
    doc.handle_data.reset_mock()

    node = parser.DataNode(data)
    node.lstrip()
    node.rstrip()
    node.write(doc)
    doc.handle_data.assert_called_once_with(both)


@pytest.mark.parametrize(
    'data',
    [
        (' '),
        ('  '),
        ('\t'),
        ('\t \t '),
    ],
)
def test_datanode_stripping_empty_string(data):
    doc = mock.Mock()
    doc.style = mock.Mock()
    doc.translate_words.side_effect = lambda words: words
    node = parser.DataNode(data)
    node.lstrip()
    node.write(doc)
    doc.handle_data.assert_not_called()


@pytest.mark.parametrize(
    'html, expected_lines',
    [
        ('<p>  foo</p>', [b'foo']),
        ('<p>\tfoo</p>', [b'foo']),
        ('<p>  <span>  </span> <span> <span> foo</span></span></p>', [b'foo']),
        ('<p>foo  </p>', [b'foo']),
        ('<p>foo\t</p>', [b'foo']),
        ('<p>  foo  </p>', [b'foo']),
        ('<p>  <span>foo</span>  </p>', [b'foo']),
        ('<p>  <span>foo  </span>  </p>', [b'foo']),
        ('<p>  <span>  foo</span>  </p>', [b'foo']),
        ('<p>  <span>  foo  </span>  </p>', [b'foo']),
        # various nested markup examples
        ('<i>italic</i>', [b'*italic*']),
        ('<p><i>italic</i></p>', [b'*italic*']),
        ('<p><i>italic</i> </p>', [b'*italic*']),
        ('<p><i>italic </i></p>', [b'*italic*']),
        ('<p>foo <i> italic </i> bar</p>', [b'foo *italic* bar']),
        ('<p>  <span> foo <i> bar</i> </span>  </p>', [b'foo *bar*']),
        ('<p>  <span> foo<i> bar</i> </span>  </p>', [b'foo* bar*']),
        ('<p>  <span> foo <i>bar</i> </span>  </p>', [b'foo *bar*']),
        # links
        ('<a href="url">foo</a> <i>bar</i>', [b'`foo <url>`__ *bar*']),
        # ReST does not support link text starting with whitespace
        ('<p>abc<a href="url"> foo</a></p>', [b'abc `foo <url>`__']),
        ('<p>abc<a href="url"> foo </a> bar</p>', [b'abc `foo <url>`__ bar']),
        # code-in-a removed and whitespace removed
        ('<a href="url"> <code>foo</code> </a> bar', [b'`foo <url>`__ bar']),
        # list items
        ('<li>  foo</li>', [b'* foo']),
        ('<li>  <foo>  </foo><foo> foo</foo></li>', [b'* foo']),
        ('<li>  <foo> foo</foo><foo>  </foo></li>', [b'* foo']),
        ('<li><foo>  </foo><foo> foo</foo> <foo>bar</li>', [b'* foo bar']),
        ('<li><foo>  </foo><foo> foo</foo> <foo> bar</li>', [b'* foo bar']),
        ('<li><foo>  </foo><foo> foo</foo><foo> bar</li>', [b'* foo bar']),
        # multiple block tags in sequence are each left and right stripped
        ('<p>  foo</p><p>  bar\t</p>', [b'foo', b'bar']),
        ('<p>  foo</p><li>  bar  </li>', [b'foo', b'* bar']),
        # nested block tags also work
        (
            '<p> <p> foo </p> <p> <span> bar </span> </p> </p>',
            [b'foo', b'bar'],
        ),
    ],
)
def test_whitespace_collapsing(html, expected_lines):
    docstring_parser = parser.DocStringParser(ReSTDocument())
    docstring_parser.feed(html)
    docstring_parser.close()
    actual = docstring_parser.doc.getvalue()

    # Get each line and filter out empty lines
    contents = actual.split(b'\n')
    contents = [line for line in contents if line and not line.isspace()]
    for line in expected_lines:
        assert line in contents
        beginning = contents.index(line)
        contents = contents[beginning:]



# Copyright 2017 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import contextlib
import datetime
import json

import botocore.endpoint
from botocore.config import Config
from botocore.exceptions import ClientError
from tests import BaseSessionTest, ClientHTTPStubber, mock
from tests.functional.test_useragent import (
    get_captured_ua_strings,
    parse_registered_feature_ids,
)

RETRY_MODES = ('legacy', 'standard', 'adaptive')


class BaseRetryTest(BaseSessionTest):
    def setUp(self):
        super().setUp()
        self.region = 'us-west-2'
        self.sleep_patch = mock.patch('time.sleep')
        self.sleep_patch.start()

    def tearDown(self):
        super().tearDown()
        self.sleep_patch.stop()

    @contextlib.contextmanager
    def assert_will_retry_n_times(
        self, client, num_retries, status=500, body=b'{}'
    ):
        num_responses = num_retries + 1
        if not isinstance(body, bytes):
            body = json.dumps(body).encode()
        with ClientHTTPStubber(client) as http_stubber:
            for _ in range(num_responses):
                http_stubber.add_response(status=status, body=body)
            with self.assertRaisesRegex(
                ClientError, f'reached max retries: {num_retries}'
            ):
                yield
            self.assertEqual(len(http_stubber.requests), num_responses)

    def _get_feature_id_lists_from_retries(self, client):
        with ClientHTTPStubber(client) as http_stubber:
            # Add two failed responses followed by a success
            http_stubber.add_response(status=502, body=b'{}')
            http_stubber.add_response(status=502, body=b'{}')
            http_stubber.add_response(status=200, body=b'{}')
            client.list_tables()
        ua_strings = get_captured_ua_strings(http_stubber)
        return [
            parse_registered_feature_ids(ua_string) for ua_string in ua_strings
        ]


class TestRetryHeader(BaseRetryTest):
    def _retry_headers_test_cases(self):
        responses = [
            [
                (500, {'Date': 'Sat, 01 Jun 2019 00:00:00 GMT'}),
                (500, {'Date': 'Sat, 01 Jun 2019 00:00:01 GMT'}),
                (200, {'Date': 'Sat, 01 Jun 2019 00:00:02 GMT'}),
            ],
            [
                (500, {'Date': 'Sat, 01 Jun 2019 00:10:03 GMT'}),
                (500, {'Date': 'Sat, 01 Jun 2019 00:10:09 GMT'}),
                (200, {'Date': 'Sat, 01 Jun 2019 00:10:15 GMT'}),
            ],
        ]

        # The first, third and seventh datetime values of each
        # utcnow_side_effects list are side_effect values for when
        # utcnow is called in SigV4 signing.
        utcnow_side_effects = [
            [
                datetime.datetime(2019, 6, 1, 0, 0, 0, 0),
                datetime.datetime(2019, 6, 1, 0, 0, 0, 0),
                datetime.datetime(2019, 6, 1, 0, 0, 1, 0),
                datetime.datetime(2019, 6, 1, 0, 0, 0, 0),
                datetime.datetime(2019, 6, 1, 0, 0, 1, 0),
                datetime.datetime(2019, 6, 1, 0, 0, 2, 0),
                datetime.datetime(2019, 6, 1, 0, 0, 0, 0),
            ],
            [
                datetime.datetime(2020, 6, 1, 0, 0, 0, 0),
                datetime.datetime(2019, 6, 1, 0, 0, 5, 0),
                datetime.datetime(2019, 6, 1, 0, 0, 6, 0),
                datetime.datetime(2019, 6, 1, 0, 0, 0, 0),
                datetime.datetime(2019, 6, 1, 0, 0, 11, 0),
                datetime.datetime(2019, 6, 1, 0, 0, 12, 0),
                datetime.datetime(2019, 6, 1, 0, 0, 0, 0),
            ],
        ]
        expected_headers = [
            [
                b'attempt=1',
                b'ttl=20190601T000011Z; attempt=2; max=3',
                b'ttl=20190601T000012Z; attempt=3; max=3',
            ],
            [
                b'attempt=1',
                b'ttl=20190601T001014Z; attempt=2; max=3',
                b'ttl=20190601T001020Z; attempt=3; max=3',
            ],
        ]
        test_cases = list(
            zip(responses, utcnow_side_effects, expected_headers)
        )
        return test_cases

    def _test_amz_sdk_request_header_with_test_case(
        self, responses, utcnow_side_effects, expected_headers, client_config
    ):
        datetime_patcher = mock.patch.object(
            botocore.endpoint.datetime,
            'datetime',
            mock.Mock(wraps=datetime.datetime),
        )
        mocked_datetime = datetime_patcher.start()
        mocked_datetime.now.side_effect = utcnow_side_effects

        client = self.session.create_client(
            'dynamodb', self.region, config=client_config
        )
        with ClientHTTPStubber(client) as http_stubber:
            for response in responses:
                http_stubber.add_response(
                    headers=response[1], status=response[0], body=b'{}'
                )
            client.list_tables()
            amz_sdk_request_headers = [
                request.headers['amz-sdk-request']
                for request in http_stubber.requests
            ]
            self.assertListEqual(amz_sdk_request_headers, expected_headers)
        datetime_patcher.stop()

    def test_amz_sdk_request_header(self):
        test_cases = self._retry_headers_test_cases()
        for retry_mode in RETRY_MODES:
            retries_config = {'mode': retry_mode, 'total_max_attempts': 3}
            client_config = Config(read_timeout=10, retries=retries_config)
            for test_case in test_cases:
                self._test_amz_sdk_request_header_with_test_case(
                    *test_case, client_config=client_config
                )

    def test_amz_sdk_invocation_id_header_persists(self):
        for retry_mode in RETRY_MODES:
            client_config = Config(retries={'mode': retry_mode})
            client = self.session.create_client(
                'dynamodb', self.region, config=client_config
            )
            num_retries = 2
            with ClientHTTPStubber(client) as http_stubber:
                for _ in range(num_retries):
                    http_stubber.add_response(status=500)
                http_stubber.add_response(status=200)
                client.list_tables()
                amz_sdk_invocation_id_headers = [
                    request.headers['amz-sdk-invocation-id']
                    for request in http_stubber.requests
                ]
                self.assertEqual(
                    amz_sdk_invocation_id_headers[0],
                    amz_sdk_invocation_id_headers[1],
                )
                self.assertEqual(
                    amz_sdk_invocation_id_headers[1],
                    amz_sdk_invocation_id_headers[2],
                )

    def test_amz_sdk_invocation_id_header_unique_per_invocation(self):
        client = self.session.create_client('dynamodb', self.region)
        num_of_invocations = 2
        with ClientHTTPStubber(client) as http_stubber:
            for _ in range(num_of_invocations):
                http_stubber.add_response(status=500)
                http_stubber.add_response(status=200)
                client.list_tables()
            amz_sdk_invocation_id_headers = [
                request.headers['amz-sdk-invocation-id']
                for request in http_stubber.requests
            ]
            self.assertEqual(
                amz_sdk_invocation_id_headers[0],
                amz_sdk_invocation_id_headers[1],
            )
            self.assertEqual(
                amz_sdk_invocation_id_headers[2],
                amz_sdk_invocation_id_headers[3],
            )
            self.assertNotEqual(
                amz_sdk_invocation_id_headers[0],
                amz_sdk_invocation_id_headers[2],
            )


class TestLegacyRetry(BaseRetryTest):
    def test_can_override_max_attempts(self):
        client = self.session.create_client(
            'dynamodb', self.region, config=Config(retries={'max_attempts': 1})
        )
        with self.assert_will_retry_n_times(client, 1):
            client.list_tables()

    def test_do_not_attempt_retries(self):
        client = self.session.create_client(
            'dynamodb', self.region, config=Config(retries={'max_attempts': 0})
        )
        with self.assert_will_retry_n_times(client, 0):
            client.list_tables()

    def test_setting_max_attempts_does_not_set_for_other_clients(self):
        # Make one client with max attempts configured.
        self.session.create_client(
            'codecommit',
            self.region,
            config=Config(retries={'max_attempts': 1}),
        )

        # Make another client that has no custom retry configured.
        client = self.session.create_client('codecommit', self.region)
        # It should use the default max retries, which should be four retries
        # for this service.
        with self.assert_will_retry_n_times(client, 4):
            client.list_repositories()

    def test_service_specific_defaults_do_not_mutate_general_defaults(self):
        # This tests for a bug where if you created a client for a service
        # with specific retry configurations and then created a client for
        # a service whose retry configurations fallback to the general
        # defaults, the second client would actually use the defaults of
        # the first client.

        # Make a dynamodb client. It's a special case client that is
        # configured to a make a maximum of 10 requests (9 retries).
        client = self.session.create_client('dynamodb', self.region)
        with self.assert_will_retry_n_times(client, 9):
            client.list_tables()

        # A codecommit client is not a special case for retries. It will at
        # most make 5 requests (4 retries) for its default.
        client = self.session.create_client('codecommit', self.region)
        with self.assert_will_retry_n_times(client, 4):
            client.list_repositories()

    def test_set_max_attempts_on_session(self):
        self.session.set_default_client_config(
            Config(retries={'max_attempts': 1})
        )
        # Max attempts should be inherited from the session.
        client = self.session.create_client('codecommit', self.region)
        with self.assert_will_retry_n_times(client, 1):
            client.list_repositories()

    def test_can_clobber_max_attempts_on_session(self):
        self.session.set_default_client_config(
            Config(retries={'max_attempts': 1})
        )
        # Max attempts should override the session's configured max attempts.
        client = self.session.create_client(
            'codecommit',
            self.region,
            config=Config(retries={'max_attempts': 0}),
        )
        with self.assert_will_retry_n_times(client, 0):
            client.list_repositories()

    def test_user_agent_has_legacy_mode_feature_id(self):
        client = self.session.create_client('dynamodb', self.region)
        feature_lists = self._get_feature_id_lists_from_retries(client)
        # Confirm all requests register `'RETRY_MODE_LEGACY': 'D'`
        assert all('D' in feature_list for feature_list in feature_lists)


class TestRetriesV2(BaseRetryTest):
    def create_client_with_retry_mode(
        self, service, retry_mode, max_attempts=None
    ):
        retries = {'mode': retry_mode}
        if max_attempts is not None:
            retries['total_max_attempts'] = max_attempts
        client = self.session.create_client(
            service, self.region, config=Config(retries=retries)
        )
        return client

    def test_standard_mode_has_default_3_retries(self):
        client = self.create_client_with_retry_mode(
            'dynamodb', retry_mode='standard'
        )
        with self.assert_will_retry_n_times(client, 2):
            client.list_tables()

    def test_standard_mode_can_configure_max_attempts(self):
        client = self.create_client_with_retry_mode(
            'dynamodb', retry_mode='standard', max_attempts=5
        )
        with self.assert_will_retry_n_times(client, 4):
            client.list_tables()

    def test_no_retry_needed_standard_mode(self):
        client = self.create_client_with_retry_mode(
            'dynamodb', retry_mode='standard'
        )
        with ClientHTTPStubber(client) as http_stubber:
            http_stubber.add_response(status=200, body=b'{}')
            client.list_tables()

    def test_standard_mode_retry_throttling_error(self):
        client = self.create_client_with_retry_mode(
            'dynamodb', retry_mode='standard'
        )
        error_body = {"__type": "ThrottlingException", "message": "Error"}
        with self.assert_will_retry_n_times(
            client, 2, status=400, body=error_body
        ):
            client.list_tables()

    def test_standard_mode_retry_transient_error(self):
        client = self.create_client_with_retry_mode(
            'dynamodb', retry_mode='standard'
        )
        with self.assert_will_retry_n_times(client, 2, status=502):
            client.list_tables()

    def test_adaptive_mode_still_retries_errors(self):
        # Verify that adaptive mode is just adding on to standard mode.
        client = self.create_client_with_retry_mode(
            'dynamodb', retry_mode='adaptive'
        )
        with self.assert_will_retry_n_times(client, 2):
            client.list_tables()

    def test_adaptive_mode_retry_transient_error(self):
        client = self.create_client_with_retry_mode(
            'dynamodb', retry_mode='adaptive'
        )
        with self.assert_will_retry_n_times(client, 2, status=502):
            client.list_tables()

    def test_can_exhaust_default_retry_quota(self):
        # Quota of 500 / 5 retry costs == 100 retry attempts
        # 100 retry attempts / 2 retries per API call == 50 client calls
        client = self.create_client_with_retry_mode(
            'dynamodb', retry_mode='standard'
        )
        for i in range(50):
            with self.assert_will_retry_n_times(client, 2, status=502):
                client.list_tables()
        # Now on the 51th attempt we should see quota errors, which we can
        # verify by looking at the request metadata.
        with ClientHTTPStubber(client) as http_stubber:
            http_stubber.add_response(status=502, body=b'{}')
            with self.assertRaises(ClientError) as e:
                client.list_tables()
        self.assertTrue(
            e.exception.response['ResponseMetadata'].get('RetryQuotaReached')
        )

    def test_user_agent_has_standard_mode_feature_id(self):
        client = self.create_client_with_retry_mode(
            'dynamodb', retry_mode='standard'
        )
        feature_lists = self._get_feature_id_lists_from_retries(client)
        # Confirm all requests register `'RETRY_MODE_STANDARD': 'E'`
        assert all('E' in feature_list for feature_list in feature_lists)

    def test_user_agent_has_adaptive_mode_feature_id(self):
        client = self.create_client_with_retry_mode(
            'dynamodb', retry_mode='adaptive'
        )
        feature_lists = self._get_feature_id_lists_from_retries(client)
        # Confirm all requests register `'RETRY_MODE_ADAPTIVE': 'F'`
        assert all('F' in feature_list for feature_list in feature_lists)
